/* pcap.c
 * main function for dissect a pcap file
 *
 * $Id: pcap.c,v 1.12 2007/09/08 07:05:41 costa Exp $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <pcap.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>

#include "log.h"
#include "packet.h"
#include "dmemory.h"
#include "proto.h"
#include "flow.h"
#include "pcapd.h"
#include "report.h"

static int pcap_prot_id;
static unsigned long pkt_serial = 0; 


static int ListSort(const void *a, const void *b)
{
    return strcmp(*(char **)a, *(char **)b);
}


static int PcapParam(int argc, char *argv[], char *file, char *dir)
{
    int c;
    short n;
    extern char *optarg;
    extern int optind, optopt;

    n = 0;
    while ((c = getopt(argc, argv, "f:d:")) != -1) {
        switch(c) {
        case 'f':
            strcpy(file, optarg);
            n++;
            break;

        case 'd':
            strcpy(dir, optarg);
            n++;
            break;

        case '?':
            printf("Error: unrecognized option: -%c\n", optopt);
            return -1;
        }
    }
    if (n != 1)
        return -1;
    return 0;
}


static void PcapDissector(u_char *user, const struct pcap_pkthdr *h, const u_char *bytes)
{
    struct pcap_ref *ref = (struct pcap_ref *)user;
    packet *pkt;
    static time_t tm = 0;
    unsigned long len;

    pkt = PktNew();

    ref->cnt++;
    pkt->raw = DMemMalloc(h->caplen+sizeof(unsigned long)*2+sizeof(char *)+4);
    memcpy(pkt->raw, bytes, h->caplen);
    pkt->raw_len = h->caplen;
    /* align 4b */
    len = pkt->raw_len;
    len = len + 4 - (len%4);
    *((unsigned long *)&(pkt->raw[len])) = ref->dlt;
    *((unsigned long *)&(pkt->raw[len+sizeof(unsigned long)])) = ref->cnt;
    *((char **)(&(pkt->raw[len+sizeof(unsigned long)*2]))) = ref->file_name;
    pkt->cap_sec = h->ts.tv_sec;
    pkt->cap_usec = h->ts.tv_usec;
    pkt->serial = pkt_serial;
    FlowSetGblTime(h->ts.tv_sec);

    ProtDissec(pcap_prot_id, pkt);

    /* next serial number */
    pkt_serial++;
    if (time(NULL) > tm) {
        tm = time(NULL) + 10;
        ReportSplash();
    }
}


char* CaptDisOptions(void)
{
    return "{-f <file.pcap> | -d <pcap_dir_path>}";
}


void CaptDisOptionsHelp(void)
{
    printf("\t-f pcap file name\n");
    printf("\t-d pcap files dir path\n");
}


int CaptDisMain(int argc, char *argv[])
{
    char errbuf[PCAP_ERRBUF_SIZE];
    char infile[PCP_PATH_DIM], dirpath[PCP_PATH_DIM];
    char **list;
    char *pcap_file;
    int i, num;
    pcap_t *cap = NULL;
    int ret;
    DIR *dir;
    struct dirent *entry;
    struct pcap_ref ref;
            
    /* pcapfile  protocol id */
    pcap_prot_id = ProtId("pcapf");
    if (pcap_prot_id == -1) {
        return -1;
    }

    /* serial number of packet */
    pkt_serial = 1;

    /* pcap file/dir name */
    infile[0] = '\0';
    dirpath[0] = '\0';
    ret = PcapParam(argc, argv, infile, dirpath);
    if (ret != 0) {
        return -1;
    }

    list = NULL;
    num = 0;
    if (dirpath[0] != '\0') {
        dir = opendir(dirpath);
        if (dir == NULL) {
            perror("");
            return -1;
        }
        /* file list */
        while((entry = readdir(dir)) != NULL) {
            if (entry->d_name[0] == '.')
                continue;
            list = xrealloc(list, sizeof(char *)*(num+1));
            list[num] = xmalloc(strlen(dirpath)+strlen(entry->d_name)+5);
            sprintf(list[num], "%s/%s", dirpath, entry->d_name);
            num++;
        }
        qsort(list, num, sizeof(char *), ListSort);
        closedir(dir);
        if (num == 0) {
            printf("Directory without pcap file\n");
            return -1;
        }
        pcap_file = list[0];
    }
    else
        pcap_file = infile;

    i = 0;
    do {
        errbuf[sizeof(errbuf) - 1] = '\0';
        errbuf[0] = '\0';
        
        /* open the input pcap file */
        cap = pcap_open_offline(pcap_file, errbuf);
        if (cap == NULL) {
            printf("File %s: %s\n", pcap_file, errbuf);
        }
        else {
            /* file name */
            ref.file_name = pcap_file;

            /* data link type */
            ref.dlt = pcap_datalink(cap);
            
            /* packet counter */
            ref.cnt = 0;

            /* let pcap loop over the input, passing data to the decryptor */
            pcap_loop(cap, -1, (pcap_handler)&PcapDissector, (u_char*)&ref);
            
            pcap_close(cap);
        }
        i++;
        if (i < num)
            pcap_file = list[i];
    } while (i<num);

    return 0;
}
