/* dns.c
 * dns packet dissection
 * RFC 1034, RFC 1035
 *
 * $Id: $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 * based on: ettercap -- dissector DNS -- UDP 53
 *   Copyright ALoR & NaGA. Web http://ettercap.sourceforge.net/
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <pcap.h>
#include <stdio.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include "etypes.h"
#include "proto.h"
#include "dmemory.h"
#include "log.h"
#include "dnsdb.h"
#include "dns.h"


/* info id */
static int ip_id;
static int ipv6_id;
static int ip_src_id;
static int ip_dst_id;
static int ipv6_src_id;
static int ipv6_dst_id;
static int tcp_id;
static int tcp_port_src_id;
static int tcp_port_dst_id;
static int tcp_lost_id;
static int tcp_clnt_id;
static int udp_id;
static int udp_port_src_id;
static int udp_port_dst_id;
static int dns_id;

static volatile unsigned int incr;


static packet* DnsDissector(int flow_id)
{
     packet *pkt;
     dns_hdr *dns_h;
     char name[NS_MAXDNAME], dummy[NS_MAXDNAME];
     unsigned char *data, *end, *nxt;
     int name_len;
     short class;
     unsigned short type, i, dim, rdlen;
     ftval ip, d_name;

     LogPrintf(LV_DEBUG, "DNS id: %d", flow_id);

     pkt = FlowGetPkt(flow_id);
     while (pkt != NULL) {
         if (pkt->len > sizeof(dns_hdr)) {
             dns_h = (dns_hdr *)pkt->data;
             /* we are interested only in DNS answer */
             if (dns_h->qr && dns_h->num_answer != 0 && dns_h->opcode == ns_o_query && dns_h->rcode == ns_r_noerror) {
                 /* extract the name from the packet */
                 data = (unsigned char *)(dns_h + 1);
                 end = (unsigned char *)(pkt->data + pkt->len);
                 name_len = dn_expand((unsigned char *)pkt->data, end, data, name, sizeof(name));
                 nxt = data + name_len;
                 type = htons(*(unsigned short *)nxt);
                 nxt += 2;
                 class = htons(*(unsigned short *)nxt);
                 nxt += 2;
                 /* handle only internet class */
                 if (class == ns_c_in) {
                     /* it is and address resolution (name to ip) */
                     if (type == ns_t_a || type == ns_t_aaaa) {
                         //ProtStackFrmDisp(pkt->stk, TRUE);
                         dim = htons(dns_h->num_q);
                         for (i=1; i<dim; i++) { /* first it was decoded */
                             if ((*nxt & NS_CMPRSFLGS) == NS_CMPRSFLGS) {
                                 nxt += 2;
                             }
                             else {
                                 name_len = dn_expand((unsigned char *)pkt->data, end, nxt, dummy, sizeof(dummy));
                                 nxt += name_len;
                             }
                             nxt += 4;
                         }

                         dim = htons(dns_h->num_answer);
                         for (i=0; i<dim; i++) {
                             if ((*nxt & NS_CMPRSFLGS) == NS_CMPRSFLGS) {
                                 nxt += 2;
                             }
                             else {
                                 name_len = dn_expand((unsigned char *)pkt->data, end, nxt, dummy, sizeof(dummy));
                                 nxt += name_len;
                             }
                             type = htons(*(unsigned short *)nxt);
                             nxt += 2;
                             class = htons(*(unsigned short *)nxt);
                             nxt += 2;
                             nxt += 4; /* ttl */
                             rdlen = htons(*(unsigned short *)nxt);
                             nxt += 2;
                             if (class == ns_c_in) {
                                 if (type == ns_t_a && rdlen == NS_INADDRSZ) {
                                     ip.int32 = *(unsigned long *)nxt;
                                     name_len = strlen(name);
                                     d_name.str = DMemMalloc(name_len + 1);
                                     memcpy(d_name.str, name, name_len);
                                     d_name.str[name_len] = '\0';
                                     DnsDbInset(&d_name, FT_STRING, &ip, FT_IPv4);
                                 }
                                 else if (type == ns_t_aaaa && rdlen == NS_IN6ADDRSZ) {
                                     memcpy(ip.ipv6, nxt, NS_IN6ADDRSZ);
                                     name_len = strlen(name);
                                     d_name.str = DMemMalloc(name_len + 1);
                                     memcpy(d_name.str, name, name_len);
                                     d_name.str[name_len] = '\0';
                                     DnsDbInset(&d_name, FT_STRING, &ip, FT_IPv6);
                                 }
                             }
                             nxt += rdlen;
                         }
                     }
                     else if (type == ns_t_ptr) {
                         /* it is a reverse query (ip to name) */
                     }
                 }
             }
         }
         
         /* new packet */
         PktFree(pkt);
         pkt = FlowGetPkt(flow_id);
     }

     return NULL;
}


static bool DnsVerifyCheck(int flow_id, bool check)
{
    packet *pkt;
    bool udp;
    bool ret;
    dns_hdr *dns_h;
    short cnt, cnt_lim;

    cnt = 0;
    udp = FALSE;
    ret = !check;
    pkt = FlowGetPktCp(flow_id);

    /* numer of packet to verify */
    if (FlowIsClose(flow_id) == TRUE)
        cnt_lim = FlowPktNum(flow_id);
    else
        cnt_lim = DNS_PKT_VER_LIMIT - 1;

    if (pkt != NULL) {
        do {
            /* verify header and data... where posible */
            if (pkt->len < sizeof(dns_hdr))
                break;
            dns_h = (dns_hdr *)pkt->data;
            if (dns_h->opcode > 2)
                break;
            if (dns_h->unused != 0)
                break;
            if (dns_h->rcode > 5)
                break;
            cnt++;
            if (check == FALSE && cnt == cnt_lim) {
                ret = TRUE;
                break;
            }
            else {
                if (dns_h->qr == 0) {
                    if (dns_h->num_answer != 0 || dns_h->num_q == 0)
                        break;
                }
                else if (dns_h->num_q == 0) {
                    break;
                }
                if (cnt == cnt_lim) {
                    ret = TRUE;
                    break;
                }
            }
            PktFree(pkt);
            pkt = FlowGetPktCp(flow_id);
        } while (pkt != NULL);
    }
    
    if (pkt != NULL)
        PktFree(pkt);

    return ret;
}


static bool DnsVerify(int flow_id)
{
    return DnsVerifyCheck(flow_id, FALSE);
}


static bool DnsCheck(int flow_id)
{
    return DnsVerifyCheck(flow_id, TRUE);
}


int DissecRegist(void)
{
    proto_dep dep;
    proto_heury_dep hdep;

    memset(&dep, 0, sizeof(proto_dep));
    memset(&hdep, 0, sizeof(proto_heury_dep));

    /* protocol name */
    ProtName("Domain Name Service", "dns");
    
    /* dep: tcp */
    dep.name = "udp";
    dep.attr = "udp.dstport";
    dep.type = FT_UINT16;
    dep.val.uint16 = TCP_PORT_DNS;
    dep.ProtCheck = DnsVerify;
    dep.pktlim = DNS_PKT_VER_LIMIT;
    ProtDep(&dep);

    /* dep: tcp */
    dep.name = "tcp";
    dep.attr = "tcp.dstport";
    dep.type = FT_UINT16;
    dep.val.uint16 = TCP_PORT_DNS;
    dep.ProtCheck = DnsVerify;
    dep.pktlim = DNS_PKT_VER_LIMIT;
    ProtDep(&dep);

    /* hdep: udp */
    hdep.name = "udp";
    hdep.ProtCheck = DnsCheck;
    hdep.pktlim = DNS_PKT_VER_LIMIT;
    ProtHeuDep(&hdep);

    /* hdep: tcp */
    hdep.name = "tcp";
    hdep.ProtCheck = DnsCheck;
    ProtHeuDep(&hdep);
    
    /* dissectors registration */
    ProtDissectors(NULL, DnsDissector, NULL, NULL);

    return 0;
}


int DissectInit(void)
{
    /* part of file name */
    incr = 0;
    
    /* protocols and attributes */
    ip_id = ProtId("ip");
    ip_dst_id = ProtAttrId(ip_id, "ip.dst");
    ip_src_id = ProtAttrId(ip_id, "ip.src");
    ipv6_id = ProtId("ipv6");
    ipv6_dst_id = ProtAttrId(ipv6_id, "ipv6.dst");
    ipv6_src_id = ProtAttrId(ipv6_id, "ipv6.src");
    tcp_id = ProtId("tcp");
    tcp_port_dst_id = ProtAttrId(tcp_id, "tcp.dstport");
    tcp_port_src_id = ProtAttrId(tcp_id, "tcp.srcport");
    tcp_lost_id = ProtAttrId(tcp_id, "tcp.lost");
    tcp_clnt_id = ProtAttrId(tcp_id, "tcp.clnt");
    udp_id = ProtId("udp");
    udp_port_dst_id = ProtAttrId(tcp_id, "udp.dstport");
    udp_port_src_id = ProtAttrId(tcp_id, "udp.srcport");
    dns_id = ProtId("dns");

    return 0;
}
