/* udp_garbage.c
 * Dissector to group together packet of udp flow that haven't a specific dissector
 *
 * $Id: udp_garbage.c,v 1.2 2007/09/08 07:13:16 costa Exp $
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <pcap.h>


#include "proto.h"
#include "dmemory.h"
#include "etypes.h"
#include "flow.h"
#include "log.h"
#include "udp_garbage.h"


static int ppp_id;
static int eth_id;
static int ip_id;
static int ipv6_id;
static int udp_id;
static int ip_src_id;
static int ip_dst_id;
static int port_src_id;
static int port_dst_id;
static volatile int serial = 0;

#define GRB_FILE     0  /* to put (or not) data in to a file */

static bool UdpGrbCheck(int flow_id)
{
    if (FlowPktNum(flow_id) >  UDP_GRB_PKT_LIMIT || FlowIsClose(flow_id) == TRUE) {
        return TRUE;
    }

    return FALSE;
}


packet* UdpGrbDissector(int flow_id)
{
    packet* pkt;
    const pstack_f *udp, *ip;
    ftval port_src, port_dst, ip_src, ip_dst;
    struct in_addr ip_addr;
    int fd_pcap;
    char filename[256], ips_str[30], ipd_str[30];
    struct pcap_file_header fh;
    struct pcap_pkthdr pckt_header;
    int count;

    LogPrintf(LV_DEBUG, "UDP garbage id: %d", flow_id);

    udp = FlowStack(flow_id);
    ip = ProtGetNxtFrame(udp);
    ProtGetAttr(udp, port_src_id, &port_src);
    ProtGetAttr(udp, port_dst_id, &port_dst);
    
    if (ProtFrameProtocol(ip) == ip_id) {
        ProtGetAttr(ip, ip_src_id, &ip_src);
        ProtGetAttr(ip, ip_dst_id, &ip_dst);
        ip_addr.s_addr = ip_src.uint32;
        LogPrintf(LV_DEBUG, "\tSRC: %s:%d", inet_ntoa(ip_addr), port_src.uint16);
        sprintf(ips_str, "%s.%d", inet_ntoa(ip_addr), port_src.uint16);
        ip_addr.s_addr = ip_dst.uint32;
        LogPrintf(LV_DEBUG, "\tDST: %s:%d", inet_ntoa(ip_addr), port_dst.uint16);
        sprintf(ipd_str, "%s.%d", inet_ntoa(ip_addr), port_dst.uint16);
    }
    else
        LogPrintf(LV_DEBUG, "UDP garbage IPv6");

    /* file pcap */
#if GRB_FILE
    sprintf(filename, "%s/udp_%d_grb_%s_%s.pcap", ProtTmpDir(), serial, ips_str, ipd_str);
#else
    sprintf(filename, "/dev/null");
#endif
    serial++;
    fd_pcap = open(filename, O_WRONLY | O_CREAT, 0x01B6);
    memset(&fh, 0, sizeof(struct pcap_file_header));
    fh.magic = 0xA1B2C3D4;
    fh.version_major = PCAP_VERSION_MAJOR;
    fh.version_minor = PCAP_VERSION_MINOR;
    fh.snaplen = 65535;
    if (ProtGetNxtFrame(ip) != NULL) {
        if (ProtFrameProtocol(ProtGetNxtFrame(ip)) == eth_id)
            fh.linktype = DLT_EN10MB;
        else
            fh.linktype = DLT_RAW;
    }
    if (fd_pcap != -1)
        write(fd_pcap, (char *)&fh, sizeof(struct pcap_file_header));

    count = 0;
    pkt = FlowGetPkt(flow_id);
    while (pkt != NULL) {
        count++;
        pckt_header.caplen = pkt->raw_len;
        pckt_header.len = pkt->raw_len;
        pckt_header.ts.tv_sec = pkt->cap_sec;
        pckt_header.ts.tv_usec = pkt->cap_usec;
        if (fd_pcap != -1) {
            write(fd_pcap, (char *)&pckt_header, sizeof(struct pcap_pkthdr));
            write(fd_pcap, (char *)pkt->raw, pkt->raw_len);
        }

        PktFree(pkt);
        pkt = FlowGetPkt(flow_id);
    }

    /* end */
    if (fd_pcap != -1)
        close(fd_pcap);

    LogPrintf(LV_DEBUG, "UDP garbage... bye bye  fid:%d", flow_id);

    return NULL;
}


int DissecRegist(void)
{
    proto_heury_dep hdep;

    memset(&hdep, 0, sizeof(proto_heury_dep));

    /* protocol name */
    ProtName("UDP garbage", "udp-grb");

    /* dep: ethernet */
    hdep.name = "udp";
    hdep.ProtCheck = UdpGrbCheck;
    ProtHeuDep(&hdep);

    /* dissectors subdissectors registration */
    ProtDissectors(NULL, UdpGrbDissector, NULL, NULL);

    return 0;
}


int DissectInit(void)
{
    ppp_id = ProtId("ppp");
    eth_id = ProtId("eth");
    ip_id = ProtId("ip");
    ipv6_id = ProtId("ipv6");
    udp_id = ProtId("udp");
    ip_dst_id = ProtAttrId(ip_id, "ip.dst");
    ip_src_id = ProtAttrId(ip_id, "ip.src");
    port_dst_id = ProtAttrId(udp_id, "udp.dstport");
    port_src_id = ProtAttrId(udp_id, "udp.srcport");

    return 0;
}
